{
 "cells": [
  {
   "cell_type": "markdown",
   "source": [
    "Surname Generation - Unconditioned\r\n",
    "===================================="
   ],
   "metadata": {}
  },
  {
   "cell_type": "markdown",
   "source": [
    "In this notebook, we use the same `Surname` dataset, but not for nationality classification. The RNN here is used as a decoder that generates sequences of characters (i.e. a generated surname) instead of a featurizer for the  classification task we have seen before.\r\n",
    "\r\n",
    "The model in this notebook is **unconditioned**: it does not observe the nationality before generating a surname. In practice, being unconditioned means the GRU does not bias its computations toward any nationality. In PyTorch code, this means we use a vector of all `0`s so that the initial hidden state vector does not\r\n",
    "contribute to the computations. In [the next notebook](https://weiliu2k.github.io/CITS4012/LSTM/Surname_Generation_Conditioned.html), the computational bias is introduced through the initial\r\n",
    "hidden vector. \r\n"
   ],
   "metadata": {}
  },
  {
   "cell_type": "markdown",
   "source": [
    "## Imports"
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 1,
   "source": [
    "import os\r\n",
    "from argparse import Namespace\r\n",
    "from collections import Counter\r\n",
    "import json\r\n",
    "import re\r\n",
    "import string\r\n",
    "\r\n",
    "import numpy as np\r\n",
    "import pandas as pd\r\n",
    "import torch\r\n",
    "import torch.nn as nn\r\n",
    "from torch.nn import functional as F\r\n",
    "import torch.optim as optim\r\n",
    "from torch.utils.data import Dataset, DataLoader\r\n",
    "from tqdm.notebook import tqdm"
   ],
   "outputs": [],
   "metadata": {}
  },
  {
   "cell_type": "markdown",
   "source": [
    "## Data Vectorization classes"
   ],
   "metadata": {}
  },
  {
   "cell_type": "markdown",
   "source": [
    "### Vocabulary"
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 4,
   "source": [
    "class Vocabulary(object):\r\n",
    "    \"\"\"Class to process text and extract vocabulary for mapping\"\"\"\r\n",
    "\r\n",
    "    def __init__(self, token_to_idx=None):\r\n",
    "        \"\"\"\r\n",
    "        Args:\r\n",
    "            token_to_idx (dict): a pre-existing map of tokens to indices\r\n",
    "        \"\"\"\r\n",
    "\r\n",
    "        if token_to_idx is None:\r\n",
    "            token_to_idx = {}\r\n",
    "        self._token_to_idx = token_to_idx\r\n",
    "\r\n",
    "        self._idx_to_token = {idx: token \r\n",
    "                              for token, idx in self._token_to_idx.items()}\r\n",
    "        \r\n",
    "    def to_serializable(self):\r\n",
    "        \"\"\" returns a dictionary that can be serialized \"\"\"\r\n",
    "        return {'token_to_idx': self._token_to_idx}\r\n",
    "\r\n",
    "    @classmethod\r\n",
    "    def from_serializable(cls, contents):\r\n",
    "        \"\"\" instantiates the Vocabulary from a serialized dictionary \"\"\"\r\n",
    "        return cls(**contents)\r\n",
    "\r\n",
    "    def add_token(self, token):\r\n",
    "        \"\"\"Update mapping dicts based on the token.\r\n",
    "\r\n",
    "        Args:\r\n",
    "            token (str): the item to add into the Vocabulary\r\n",
    "        Returns:\r\n",
    "            index (int): the integer corresponding to the token\r\n",
    "        \"\"\"\r\n",
    "        if token in self._token_to_idx:\r\n",
    "            index = self._token_to_idx[token]\r\n",
    "        else:\r\n",
    "            index = len(self._token_to_idx)\r\n",
    "            self._token_to_idx[token] = index\r\n",
    "            self._idx_to_token[index] = token\r\n",
    "        return index\r\n",
    "            \r\n",
    "    def add_many(self, tokens):\r\n",
    "        \"\"\"Add a list of tokens into the Vocabulary\r\n",
    "        \r\n",
    "        Args:\r\n",
    "            tokens (list): a list of string tokens\r\n",
    "        Returns:\r\n",
    "            indices (list): a list of indices corresponding to the tokens\r\n",
    "        \"\"\"\r\n",
    "        return [self.add_token(token) for token in tokens]\r\n",
    "\r\n",
    "    def lookup_token(self, token):\r\n",
    "        \"\"\"Retrieve the index associated with the token \r\n",
    "        \r\n",
    "        Args:\r\n",
    "            token (str): the token to look up \r\n",
    "        Returns:\r\n",
    "            index (int): the index corresponding to the token\r\n",
    "        \"\"\"\r\n",
    "        return self._token_to_idx[token]\r\n",
    "\r\n",
    "    def lookup_index(self, index):\r\n",
    "        \"\"\"Return the token associated with the index\r\n",
    "        \r\n",
    "        Args: \r\n",
    "            index (int): the index to look up\r\n",
    "        Returns:\r\n",
    "            token (str): the token corresponding to the index\r\n",
    "        Raises:\r\n",
    "            KeyError: if the index is not in the Vocabulary\r\n",
    "        \"\"\"\r\n",
    "        if index not in self._idx_to_token:\r\n",
    "            raise KeyError(\"the index (%d) is not in the Vocabulary\" % index)\r\n",
    "        return self._idx_to_token[index]\r\n",
    "\r\n",
    "    def __str__(self):\r\n",
    "        return \"<Vocabulary(size=%d)>\" % len(self)\r\n",
    "\r\n",
    "    def __len__(self):\r\n",
    "        return len(self._token_to_idx)"
   ],
   "outputs": [],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 5,
   "source": [
    "class SequenceVocabulary(Vocabulary):\r\n",
    "    def __init__(self, token_to_idx=None, unk_token=\"<UNK>\",\r\n",
    "                 mask_token=\"<MASK>\", begin_seq_token=\"<BEGIN>\",\r\n",
    "                 end_seq_token=\"<END>\"):\r\n",
    "\r\n",
    "        super(SequenceVocabulary, self).__init__(token_to_idx)\r\n",
    "\r\n",
    "        self._mask_token = mask_token\r\n",
    "        self._unk_token = unk_token\r\n",
    "        self._begin_seq_token = begin_seq_token\r\n",
    "        self._end_seq_token = end_seq_token\r\n",
    "\r\n",
    "        self.mask_index = self.add_token(self._mask_token)\r\n",
    "        self.unk_index = self.add_token(self._unk_token)\r\n",
    "        self.begin_seq_index = self.add_token(self._begin_seq_token)\r\n",
    "        self.end_seq_index = self.add_token(self._end_seq_token)\r\n",
    "\r\n",
    "    def to_serializable(self):\r\n",
    "        contents = super(SequenceVocabulary, self).to_serializable()\r\n",
    "        contents.update({'unk_token': self._unk_token,\r\n",
    "                         'mask_token': self._mask_token,\r\n",
    "                         'begin_seq_token': self._begin_seq_token,\r\n",
    "                         'end_seq_token': self._end_seq_token})\r\n",
    "        return contents\r\n",
    "\r\n",
    "    def lookup_token(self, token):\r\n",
    "        \"\"\"Retrieve the index associated with the token \r\n",
    "          or the UNK index if token isn't present.\r\n",
    "        \r\n",
    "        Args:\r\n",
    "            token (str): the token to look up \r\n",
    "        Returns:\r\n",
    "            index (int): the index corresponding to the token\r\n",
    "        Notes:\r\n",
    "            `unk_index` needs to be >=0 (having been added into the Vocabulary) \r\n",
    "              for the UNK functionality \r\n",
    "        \"\"\"\r\n",
    "        if self.unk_index >= 0:\r\n",
    "            return self._token_to_idx.get(token, self.unk_index)\r\n",
    "        else:\r\n",
    "            return self._token_to_idx[token]"
   ],
   "outputs": [],
   "metadata": {}
  },
  {
   "cell_type": "markdown",
   "source": [
    "### Vectorizer"
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 6,
   "source": [
    "class SurnameVectorizer(object):\r\n",
    "    \"\"\" The Vectorizer which coordinates the Vocabularies and puts them to use\"\"\"    \r\n",
    "    def __init__(self, char_vocab, nationality_vocab):\r\n",
    "        \"\"\"\r\n",
    "        Args:\r\n",
    "            char_vocab (Vocabulary): maps words to integers\r\n",
    "            nationality_vocab (Vocabulary): maps nationalities to integers\r\n",
    "        \"\"\"\r\n",
    "        self.char_vocab = char_vocab\r\n",
    "        self.nationality_vocab = nationality_vocab\r\n",
    "\r\n",
    "    def vectorize(self, surname, vector_length=-1):\r\n",
    "        \"\"\"Vectorize a surname into a vector of observations and targets\r\n",
    "        \r\n",
    "        The outputs are the vectorized surname split into two vectors:\r\n",
    "            surname[:-1] and surname[1:]\r\n",
    "        At each timestep, the first vector is the observation and the second vector is the target. \r\n",
    "        \r\n",
    "        Args:\r\n",
    "            surname (str): the surname to be vectorized\r\n",
    "            vector_length (int): an argument for forcing the length of index vector\r\n",
    "        Returns:\r\n",
    "            a tuple: (from_vector, to_vector)\r\n",
    "            from_vector (numpy.ndarray): the observation vector \r\n",
    "            to_vector (numpy.ndarray): the target prediction vector\r\n",
    "        \"\"\"\r\n",
    "        indices = [self.char_vocab.begin_seq_index] \r\n",
    "        indices.extend(self.char_vocab.lookup_token(token) for token in surname)\r\n",
    "        indices.append(self.char_vocab.end_seq_index)\r\n",
    "\r\n",
    "        if vector_length < 0:\r\n",
    "            vector_length = len(indices) - 1\r\n",
    "\r\n",
    "        from_vector = np.empty(vector_length, dtype=np.int64)         \r\n",
    "        from_indices = indices[:-1]\r\n",
    "        from_vector[:len(from_indices)] = from_indices\r\n",
    "        from_vector[len(from_indices):] = self.char_vocab.mask_index\r\n",
    "\r\n",
    "        to_vector = np.empty(vector_length, dtype=np.int64)\r\n",
    "        to_indices = indices[1:]\r\n",
    "        to_vector[:len(to_indices)] = to_indices\r\n",
    "        to_vector[len(to_indices):] = self.char_vocab.mask_index\r\n",
    "        \r\n",
    "        return from_vector, to_vector\r\n",
    "\r\n",
    "    @classmethod\r\n",
    "    def from_dataframe(cls, surname_df):\r\n",
    "        \"\"\"Instantiate the vectorizer from the dataset dataframe\r\n",
    "        \r\n",
    "        Args:\r\n",
    "            surname_df (pandas.DataFrame): the surname dataset\r\n",
    "        Returns:\r\n",
    "            an instance of the SurnameVectorizer\r\n",
    "        \"\"\"\r\n",
    "        char_vocab = SequenceVocabulary()\r\n",
    "        nationality_vocab = Vocabulary()\r\n",
    "\r\n",
    "        for index, row in surname_df.iterrows():\r\n",
    "            for char in row.surname:\r\n",
    "                char_vocab.add_token(char)\r\n",
    "            nationality_vocab.add_token(row.nationality)\r\n",
    "\r\n",
    "        return cls(char_vocab, nationality_vocab)\r\n",
    "\r\n",
    "    @classmethod\r\n",
    "    def from_serializable(cls, contents):\r\n",
    "        \"\"\"Instantiate the vectorizer from saved contents\r\n",
    "        \r\n",
    "        Args:\r\n",
    "            contents (dict): a dict holding two vocabularies for this vectorizer\r\n",
    "                This dictionary is created using `vectorizer.to_serializable()`\r\n",
    "        Returns:\r\n",
    "            an instance of SurnameVectorizer\r\n",
    "        \"\"\"\r\n",
    "        char_vocab = SequenceVocabulary.from_serializable(contents['char_vocab'])\r\n",
    "        nat_vocab =  Vocabulary.from_serializable(contents['nationality_vocab'])\r\n",
    "\r\n",
    "        return cls(char_vocab=char_vocab, nationality_vocab=nat_vocab)\r\n",
    "\r\n",
    "    def to_serializable(self):\r\n",
    "        \"\"\" Returns the serializable contents \"\"\"\r\n",
    "        return {'char_vocab': self.char_vocab.to_serializable(), \r\n",
    "                'nationality_vocab': self.nationality_vocab.to_serializable()}"
   ],
   "outputs": [],
   "metadata": {}
  },
  {
   "cell_type": "markdown",
   "source": [
    "### Dataset"
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 7,
   "source": [
    "class SurnameDataset(Dataset):\r\n",
    "    def __init__(self, surname_df, vectorizer):\r\n",
    "        \"\"\"\r\n",
    "        Args:\r\n",
    "            surname_df (pandas.DataFrame): the dataset\r\n",
    "            vectorizer (SurnameVectorizer): vectorizer instatiated from dataset\r\n",
    "        \"\"\"\r\n",
    "        self.surname_df = surname_df \r\n",
    "        self._vectorizer = vectorizer\r\n",
    "\r\n",
    "        self._max_seq_length = max(map(len, self.surname_df.surname)) + 2\r\n",
    "\r\n",
    "        self.train_df = self.surname_df[self.surname_df.split=='train']\r\n",
    "        self.train_size = len(self.train_df)\r\n",
    "\r\n",
    "        self.val_df = self.surname_df[self.surname_df.split=='val']\r\n",
    "        self.validation_size = len(self.val_df)\r\n",
    "\r\n",
    "        self.test_df = self.surname_df[self.surname_df.split=='test']\r\n",
    "        self.test_size = len(self.test_df)\r\n",
    "\r\n",
    "        self._lookup_dict = {'train': (self.train_df, self.train_size), \r\n",
    "                             'val': (self.val_df, self.validation_size), \r\n",
    "                             'test': (self.test_df, self.test_size)}\r\n",
    "\r\n",
    "        self.set_split('train')\r\n",
    "        \r\n",
    "    @classmethod\r\n",
    "    def load_dataset_and_make_vectorizer(cls, surname_csv):\r\n",
    "        \"\"\"Load dataset and make a new vectorizer from scratch\r\n",
    "        \r\n",
    "        Args:\r\n",
    "            surname_csv (str): location of the dataset\r\n",
    "        Returns:\r\n",
    "            an instance of SurnameDataset\r\n",
    "        \"\"\"\r\n",
    "        \r\n",
    "        surname_df = pd.read_csv(surname_csv)\r\n",
    "        return cls(surname_df, SurnameVectorizer.from_dataframe(surname_df))\r\n",
    "        \r\n",
    "    @classmethod\r\n",
    "    def load_dataset_and_load_vectorizer(cls, surname_csv, vectorizer_filepath):\r\n",
    "        \"\"\"Load dataset and the corresponding vectorizer. \r\n",
    "        Used in the case in the vectorizer has been cached for re-use\r\n",
    "        \r\n",
    "        Args:\r\n",
    "            surname_csv (str): location of the dataset\r\n",
    "            vectorizer_filepath (str): location of the saved vectorizer\r\n",
    "        Returns:\r\n",
    "            an instance of SurnameDataset\r\n",
    "        \"\"\"\r\n",
    "        surname_df = pd.read_csv(surname_csv)\r\n",
    "        vectorizer = cls.load_vectorizer_only(vectorizer_filepath)\r\n",
    "        return cls(surname_df, vectorizer)\r\n",
    "\r\n",
    "    @staticmethod\r\n",
    "    def load_vectorizer_only(vectorizer_filepath):\r\n",
    "        \"\"\"a static method for loading the vectorizer from file\r\n",
    "        \r\n",
    "        Args:\r\n",
    "            vectorizer_filepath (str): the location of the serialized vectorizer\r\n",
    "        Returns:\r\n",
    "            an instance of SurnameVectorizer\r\n",
    "        \"\"\"\r\n",
    "        with open(vectorizer_filepath) as fp:\r\n",
    "            return SurnameVectorizer.from_serializable(json.load(fp))\r\n",
    "\r\n",
    "    def save_vectorizer(self, vectorizer_filepath):\r\n",
    "        \"\"\"saves the vectorizer to disk using json\r\n",
    "        \r\n",
    "        Args:\r\n",
    "            vectorizer_filepath (str): the location to save the vectorizer\r\n",
    "        \"\"\"\r\n",
    "        with open(vectorizer_filepath, \"w\") as fp:\r\n",
    "            json.dump(self._vectorizer.to_serializable(), fp)\r\n",
    "\r\n",
    "    def get_vectorizer(self):\r\n",
    "        \"\"\" returns the vectorizer \"\"\"\r\n",
    "        return self._vectorizer\r\n",
    "\r\n",
    "    def set_split(self, split=\"train\"):\r\n",
    "        self._target_split = split\r\n",
    "        self._target_df, self._target_size = self._lookup_dict[split]\r\n",
    "\r\n",
    "    def __len__(self):\r\n",
    "        return self._target_size\r\n",
    "\r\n",
    "    def __getitem__(self, index):\r\n",
    "        \"\"\"the primary entry point method for PyTorch datasets\r\n",
    "        \r\n",
    "        Args:\r\n",
    "            index (int): the index to the data point \r\n",
    "        Returns:\r\n",
    "            a dictionary holding the data point: (x_data, y_target, class_index)\r\n",
    "        \"\"\"\r\n",
    "        row = self._target_df.iloc[index]\r\n",
    "        \r\n",
    "        from_vector, to_vector = \\\r\n",
    "            self._vectorizer.vectorize(row.surname, self._max_seq_length)\r\n",
    "        \r\n",
    "        nationality_index = \\\r\n",
    "            self._vectorizer.nationality_vocab.lookup_token(row.nationality)\r\n",
    "\r\n",
    "        return {'x_data': from_vector, \r\n",
    "                'y_target': to_vector, \r\n",
    "                'class_index': nationality_index}\r\n",
    "\r\n",
    "    def get_num_batches(self, batch_size):\r\n",
    "        \"\"\"Given a batch size, return the number of batches in the dataset\r\n",
    "        \r\n",
    "        Args:\r\n",
    "            batch_size (int)\r\n",
    "        Returns:\r\n",
    "            number of batches in the dataset\r\n",
    "        \"\"\"\r\n",
    "        return len(self) // batch_size\r\n",
    "    \r\n",
    "def generate_batches(dataset, batch_size, shuffle=True,\r\n",
    "                     drop_last=True, device=\"cpu\"): \r\n",
    "    \"\"\"\r\n",
    "    A generator function which wraps the PyTorch DataLoader. It will \r\n",
    "      ensure each tensor is on the write device location.\r\n",
    "    \"\"\"\r\n",
    "    dataloader = DataLoader(dataset=dataset, batch_size=batch_size,\r\n",
    "                            shuffle=shuffle, drop_last=drop_last)\r\n",
    "\r\n",
    "    for data_dict in dataloader:\r\n",
    "        out_data_dict = {}\r\n",
    "        for name, tensor in data_dict.items():\r\n",
    "            out_data_dict[name] = data_dict[name].to(device)\r\n",
    "        yield out_data_dict"
   ],
   "outputs": [],
   "metadata": {}
  },
  {
   "cell_type": "markdown",
   "source": [
    "## The Model: SurnameGenerationModel\r\n",
    "\r\n",
    "The [`SurnameGenerationModel`](#the-model-surnamegenerationmodel) embeds character indices, computes\r\n",
    "their sequential state using a GRU, and computes the probability of token predictions using a Linear\r\n",
    "layer. More explicitly, the unconditioned `SurnameGenerationModel` starts with initializing an\r\n",
    "Embedding layer, a GRU, and a Linear layer. Similar to the sequence models for classification, when a matrix of integers is input to the model:\r\n",
    "\r\n",
    "- We first use a PyTorch Embedding instance, the `char_embedding`, to convert the integers to a three dimensional tensor (a sequence of vectors for each batch item);\r\n",
    "- This tensor is passed to the GRU, which computes a hidden state vector for each character in each sequence. \r\n",
    "\r\n",
    "The primary difference between the sequence classification tasks and sequence prediction\r\n",
    "tasks of this chapter is how the state vectors computed by the RNN are handled. In classification, we\r\n",
    "retrieved a single vector per batch index and performed predictions using those single vectors. In this\r\n",
    "example, we reshape our three-dimensional tensor `(N,L,F)` into a two-dimensional tensor (a matrix `(NxL,F)`) with each row represents a character (batch and sequence index) and the columns of each row are features (embedding for the character). Using this matrix and the Linear layer, we compute prediction vectors for every character. We finish the computation by reshaping the matrix back into a threedimensional tensor. Because the ordering information is preserved with reshaping operations, each batch and sequence index is still in the same position. The reason why we needed to reshape is because the Linear layer requires a matrix as input."
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 8,
   "source": [
    "class SurnameGenerationModel(nn.Module):\r\n",
    "    def __init__(self, char_embedding_size, char_vocab_size, rnn_hidden_size, \r\n",
    "                 batch_first=True, padding_idx=0, dropout_p=0.5):\r\n",
    "        \"\"\"\r\n",
    "        Args:\r\n",
    "            char_embedding_size (int): The size of the character embeddings\r\n",
    "            char_vocab_size (int): The number of characters to embed\r\n",
    "            rnn_hidden_size (int): The size of the RNN's hidden state\r\n",
    "            batch_first (bool): Informs whether the input tensors will \r\n",
    "                have batch or the sequence on the 0th dimension\r\n",
    "            padding_idx (int): The index for the tensor padding; \r\n",
    "                see torch.nn.Embedding\r\n",
    "            dropout_p (float): the probability of zeroing activations using\r\n",
    "                the dropout method.  higher means more likely to zero.\r\n",
    "        \"\"\"\r\n",
    "        super(SurnameGenerationModel, self).__init__()\r\n",
    "        \r\n",
    "        self.char_emb = nn.Embedding(num_embeddings=char_vocab_size,\r\n",
    "                                     embedding_dim=char_embedding_size,\r\n",
    "                                     padding_idx=padding_idx)\r\n",
    "\r\n",
    "        self.rnn = nn.GRU(input_size=char_embedding_size, \r\n",
    "                          hidden_size=rnn_hidden_size,\r\n",
    "                          batch_first=batch_first)\r\n",
    "        \r\n",
    "        self.fc = nn.Linear(in_features=rnn_hidden_size, \r\n",
    "                            out_features=char_vocab_size)\r\n",
    "        \r\n",
    "        self._dropout_p = dropout_p\r\n",
    "\r\n",
    "    def forward(self, x_in, apply_softmax=False):\r\n",
    "        \"\"\"The forward pass of the model\r\n",
    "        \r\n",
    "        Args:\r\n",
    "            x_in (torch.Tensor): an input data tensor. \r\n",
    "                x_in.shape should be (batch, input_dim)\r\n",
    "            apply_softmax (bool): a flag for the softmax activation\r\n",
    "                should be false if used with the Cross Entropy losses\r\n",
    "        Returns:\r\n",
    "            the resulting tensor. tensor.shape should be (batch, char_vocab_size)\r\n",
    "        \"\"\"\r\n",
    "        x_embedded = self.char_emb(x_in)\r\n",
    "\r\n",
    "        y_out, _ = self.rnn(x_embedded)\r\n",
    "\r\n",
    "        batch_size, seq_size, feat_size = y_out.shape\r\n",
    "        y_out = y_out.contiguous().view(batch_size * seq_size, feat_size)\r\n",
    "\r\n",
    "        y_out = self.fc(F.dropout(y_out, p=self._dropout_p))\r\n",
    "                         \r\n",
    "        if apply_softmax:\r\n",
    "            y_out = F.softmax(y_out, dim=1)\r\n",
    "            \r\n",
    "        new_feat_size = y_out.shape[-1]\r\n",
    "        y_out = y_out.view(batch_size, seq_size, new_feat_size)\r\n",
    "            \r\n",
    "        return y_out"
   ],
   "outputs": [],
   "metadata": {}
  },
  {
   "cell_type": "markdown",
   "source": [
    "Even though the percharacter accuracy of the predictions is a measure of model performance, it is\r\n",
    "better in this example to do qualitative evaluation by inspecting what kinds of surnames the model will generate. To do this, we write a new loop over a modified version of the steps in the `forward()` method to compute predictions at each time step and use those predictions as the input to the following time step. Please see the code in the `sample_from_model()` function.\r\n",
    "\r\n",
    "The output of the model at each time step is a prediction vector which is turned into a probability distribution using the softmax function. Using the probability distribution, we take advantage of the `torch.multinomial()` sampling function, which selects an index at a rate proportional to the probability of the index. Sampling is a stochastic procedure that produces different outputs each time."
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 9,
   "source": [
    "def sample_from_model(model, vectorizer, num_samples=1, sample_size=20, \r\n",
    "                      temperature=1.0):\r\n",
    "    \"\"\"Sample a sequence of indices from the model\r\n",
    "    \r\n",
    "    Args:\r\n",
    "        model (SurnameGenerationModel): the trained model\r\n",
    "        vectorizer (SurnameVectorizer): the corresponding vectorizer\r\n",
    "        num_samples (int): the number of samples\r\n",
    "        sample_size (int): the max length of the samples\r\n",
    "        temperature (float): accentuates or flattens \r\n",
    "            the distribution. \r\n",
    "            0.0 < temperature < 1.0 will make it peakier. \r\n",
    "            temperature > 1.0 will make it more uniform\r\n",
    "    Returns:\r\n",
    "        indices (torch.Tensor): the matrix of indices; \r\n",
    "        shape = (num_samples, sample_size)\r\n",
    "    \"\"\"\r\n",
    "    begin_seq_index = [vectorizer.char_vocab.begin_seq_index \r\n",
    "                       for _ in range(num_samples)]\r\n",
    "    begin_seq_index = torch.tensor(begin_seq_index, \r\n",
    "                                   dtype=torch.int64).unsqueeze(dim=1)\r\n",
    "    indices = [begin_seq_index]\r\n",
    "    h_t = None\r\n",
    "    \r\n",
    "    for time_step in range(sample_size):\r\n",
    "        x_t = indices[time_step]\r\n",
    "        x_emb_t = model.char_emb(x_t)\r\n",
    "        rnn_out_t, h_t = model.rnn(x_emb_t, h_t)\r\n",
    "        prediction_vector = model.fc(rnn_out_t.squeeze(dim=1))\r\n",
    "        probability_vector = F.softmax(prediction_vector / temperature, dim=1)\r\n",
    "        indices.append(torch.multinomial(probability_vector, num_samples=1))\r\n",
    "    indices = torch.stack(indices).squeeze().permute(1, 0)\r\n",
    "    return indices\r\n"
   ],
   "outputs": [],
   "metadata": {}
  },
  {
   "cell_type": "markdown",
   "source": [
    "We need to transform the sampled indices from the `sample_from_model()` function into a string for humanreadable output. The `decode_sample()` function demonstrates, to do this, we use the `SequenceVocabulary` that was used to vectorize the surnames. In creating the string, we use only the indices up to the `ENDOFSEQUENCE` index. This assumes that the model learns some sense of\r\n",
    "when surnames should end."
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "source": [
    "\r\n",
    "def decode_samples(sampled_indices, vectorizer):\r\n",
    "    \"\"\"Transform indices into the string form of a surname\r\n",
    "    \r\n",
    "    Args:\r\n",
    "        sampled_indices (torch.Tensor): the inidces from `sample_from_model`\r\n",
    "        vectorizer (SurnameVectorizer): the corresponding vectorizer\r\n",
    "    \"\"\"\r\n",
    "    decoded_surnames = []\r\n",
    "    vocab = vectorizer.char_vocab\r\n",
    "    \r\n",
    "    for sample_index in range(sampled_indices.shape[0]):\r\n",
    "        surname = \"\"\r\n",
    "        for time_step in range(sampled_indices.shape[1]):\r\n",
    "            sample_item = sampled_indices[sample_index, time_step].item()\r\n",
    "            if sample_item == vocab.begin_seq_index:\r\n",
    "                continue\r\n",
    "            elif sample_item == vocab.end_seq_index:\r\n",
    "                break\r\n",
    "            else:\r\n",
    "                surname += vocab.lookup_index(sample_item)\r\n",
    "        decoded_surnames.append(surname)\r\n",
    "    return decoded_surnames\r\n",
    "            "
   ],
   "outputs": [],
   "metadata": {}
  },
  {
   "cell_type": "markdown",
   "source": [
    "## Training Routine"
   ],
   "metadata": {}
  },
  {
   "cell_type": "markdown",
   "source": [
    "### Helper functions"
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 10,
   "source": [
    "def make_train_state(args):\r\n",
    "    return {'stop_early': False,\r\n",
    "            'early_stopping_step': 0,\r\n",
    "            'early_stopping_best_val': 1e8,\r\n",
    "            'learning_rate': args.learning_rate,\r\n",
    "            'epoch_index': 0,\r\n",
    "            'train_loss': [],\r\n",
    "            'train_acc': [],\r\n",
    "            'val_loss': [],\r\n",
    "            'val_acc': [],\r\n",
    "            'test_loss': -1,\r\n",
    "            'test_acc': -1,\r\n",
    "            'model_filename': args.model_state_file}\r\n",
    "\r\n",
    "def update_train_state(args, model, train_state):\r\n",
    "    \"\"\"Handle the training state updates.\r\n",
    "    Components:\r\n",
    "     - Early Stopping: Prevent overfitting.\r\n",
    "     - Model Checkpoint: Model is saved if the model is better\r\n",
    "    \r\n",
    "    :param args: main arguments\r\n",
    "    :param model: model to train\r\n",
    "    :param train_state: a dictionary representing the training state values\r\n",
    "    :returns:\r\n",
    "        a new train_state\r\n",
    "    \"\"\"\r\n",
    "\r\n",
    "    # Save one model at least\r\n",
    "    if train_state['epoch_index'] == 0:\r\n",
    "        torch.save(model.state_dict(), train_state['model_filename'])\r\n",
    "        train_state['stop_early'] = False\r\n",
    "\r\n",
    "    # Save model if performance improved\r\n",
    "    elif train_state['epoch_index'] >= 1:\r\n",
    "        loss_tm1, loss_t = train_state['val_loss'][-2:]\r\n",
    "         \r\n",
    "        # If loss worsened\r\n",
    "        if loss_t >= loss_tm1:\r\n",
    "            # Update step\r\n",
    "            train_state['early_stopping_step'] += 1\r\n",
    "        # Loss decreased\r\n",
    "        else:\r\n",
    "            # Save the best model\r\n",
    "            if loss_t < train_state['early_stopping_best_val']:\r\n",
    "                torch.save(model.state_dict(), train_state['model_filename'])\r\n",
    "                train_state['early_stopping_best_val'] = loss_t\r\n",
    "\r\n",
    "            # Reset early stopping step\r\n",
    "            train_state['early_stopping_step'] = 0\r\n",
    "\r\n",
    "        # Stop early ?\r\n",
    "        train_state['stop_early'] = \\\r\n",
    "            train_state['early_stopping_step'] >= args.early_stopping_criteria\r\n",
    "\r\n",
    "    return train_state\r\n",
    "\r\n",
    "def normalize_sizes(y_pred, y_true):\r\n",
    "    \"\"\"Normalize tensor sizes\r\n",
    "    \r\n",
    "    Args:\r\n",
    "        y_pred (torch.Tensor): the output of the model\r\n",
    "            If a 3-dimensional tensor, reshapes to a matrix\r\n",
    "        y_true (torch.Tensor): the target predictions\r\n",
    "            If a matrix, reshapes to be a vector\r\n",
    "    \"\"\"\r\n",
    "    if len(y_pred.size()) == 3:\r\n",
    "        y_pred = y_pred.contiguous().view(-1, y_pred.size(2))\r\n",
    "    if len(y_true.size()) == 2:\r\n",
    "        y_true = y_true.contiguous().view(-1)\r\n",
    "    return y_pred, y_true\r\n",
    "\r\n",
    "def compute_accuracy(y_pred, y_true, mask_index):\r\n",
    "    y_pred, y_true = normalize_sizes(y_pred, y_true)\r\n",
    "\r\n",
    "    _, y_pred_indices = y_pred.max(dim=1)\r\n",
    "    \r\n",
    "    correct_indices = torch.eq(y_pred_indices, y_true).float()\r\n",
    "    valid_indices = torch.ne(y_true, mask_index).float()\r\n",
    "    \r\n",
    "    n_correct = (correct_indices * valid_indices).sum().item()\r\n",
    "    n_valid = valid_indices.sum().item()\r\n",
    "\r\n",
    "    return n_correct / n_valid * 100\r\n",
    "\r\n",
    "def sequence_loss(y_pred, y_true, mask_index):\r\n",
    "    y_pred, y_true = normalize_sizes(y_pred, y_true)\r\n",
    "    return F.cross_entropy(y_pred, y_true, ignore_index=mask_index)"
   ],
   "outputs": [],
   "metadata": {}
  },
  {
   "cell_type": "markdown",
   "source": [
    "### General utilities"
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 12,
   "source": [
    "def set_seed_everywhere(seed, cuda):\r\n",
    "    np.random.seed(seed)\r\n",
    "    torch.manual_seed(seed)\r\n",
    "    if cuda:\r\n",
    "        torch.cuda.manual_seed_all(seed)\r\n",
    "\r\n",
    "def handle_dirs(dirpath):\r\n",
    "    if not os.path.exists(dirpath):\r\n",
    "        os.makedirs(dirpath)"
   ],
   "outputs": [],
   "metadata": {}
  },
  {
   "cell_type": "markdown",
   "source": [
    "### Settings and some prep work"
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 14,
   "source": [
    "args = Namespace(\r\n",
    "    # Data and Path information\r\n",
    "    surname_csv=\"../data/surnames/surnames_with_splits.csv\",\r\n",
    "    vectorizer_file=\"vectorizer.json\",\r\n",
    "    model_state_file=\"model.pth\",\r\n",
    "    save_dir=\"model_storage/model1_unconditioned_surname_generation\",\r\n",
    "    # Model hyper parameters\r\n",
    "    char_embedding_size=32,\r\n",
    "    rnn_hidden_size=32,\r\n",
    "    # Training hyper parameters\r\n",
    "    seed=1337,\r\n",
    "    learning_rate=0.001,\r\n",
    "    batch_size=128,\r\n",
    "    num_epochs=100,\r\n",
    "    early_stopping_criteria=5,\r\n",
    "    # Runtime options\r\n",
    "    catch_keyboard_interrupt=True,\r\n",
    "    cuda=True,\r\n",
    "    expand_filepaths_to_save_dir=True,\r\n",
    "    reload_from_files=False,\r\n",
    ")\r\n",
    "\r\n",
    "if args.expand_filepaths_to_save_dir:\r\n",
    "    args.vectorizer_file = os.path.join(args.save_dir,\r\n",
    "                                        args.vectorizer_file)\r\n",
    "\r\n",
    "    args.model_state_file = os.path.join(args.save_dir,\r\n",
    "                                         args.model_state_file)\r\n",
    "    \r\n",
    "    print(\"Expanded filepaths: \")\r\n",
    "    print(\"\\t{}\".format(args.vectorizer_file))\r\n",
    "    print(\"\\t{}\".format(args.model_state_file))\r\n",
    "    \r\n",
    "    \r\n",
    "# Check CUDA\r\n",
    "if not torch.cuda.is_available():\r\n",
    "    args.cuda = False\r\n",
    "\r\n",
    "args.device = torch.device(\"cuda\" if args.cuda else \"cpu\")\r\n",
    "    \r\n",
    "print(\"Using CUDA: {}\".format(args.cuda))\r\n",
    "\r\n",
    "# Set seed for reproducibility\r\n",
    "set_seed_everywhere(args.seed, args.cuda)\r\n",
    "\r\n",
    "# handle dirs\r\n",
    "handle_dirs(args.save_dir)"
   ],
   "outputs": [
    {
     "output_type": "stream",
     "name": "stdout",
     "text": [
      "Expanded filepaths: \n",
      "\tmodel_storage/model1_unconditioned_surname_generation\\vectorizer.json\n",
      "\tmodel_storage/model1_unconditioned_surname_generation\\model.pth\n",
      "Using CUDA: True\n"
     ]
    }
   ],
   "metadata": {}
  },
  {
   "cell_type": "markdown",
   "source": [
    "### Initializations"
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 15,
   "source": [
    "if args.reload_from_files:\r\n",
    "    # training from a checkpoint\r\n",
    "    dataset = SurnameDataset.load_dataset_and_load_vectorizer(args.surname_csv,\r\n",
    "                                                              args.vectorizer_file)\r\n",
    "else:\r\n",
    "    # create dataset and vectorizer\r\n",
    "    dataset = SurnameDataset.load_dataset_and_make_vectorizer(args.surname_csv)\r\n",
    "    dataset.save_vectorizer(args.vectorizer_file)\r\n",
    "\r\n",
    "vectorizer = dataset.get_vectorizer()\r\n",
    "\r\n",
    "model = SurnameGenerationModel(char_embedding_size=args.char_embedding_size,\r\n",
    "                               char_vocab_size=len(vectorizer.char_vocab),\r\n",
    "                               rnn_hidden_size=args.rnn_hidden_size,\r\n",
    "                               padding_idx=vectorizer.char_vocab.mask_index)"
   ],
   "outputs": [],
   "metadata": {}
  },
  {
   "cell_type": "markdown",
   "source": [
    "### Training loop"
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 17,
   "source": [
    "mask_index = vectorizer.char_vocab.mask_index\r\n",
    "\r\n",
    "model = model.to(args.device)\r\n",
    "\r\n",
    "\r\n",
    "optimizer = optim.Adam(model.parameters(), lr=args.learning_rate)\r\n",
    "scheduler = optim.lr_scheduler.ReduceLROnPlateau(optimizer=optimizer,\r\n",
    "                                           mode='min', factor=0.5,\r\n",
    "                                           patience=1)\r\n",
    "train_state = make_train_state(args)\r\n",
    "\r\n",
    "epoch_bar = tqdm(desc='training routine', \r\n",
    "                          total=args.num_epochs,\r\n",
    "                          position=0)\r\n",
    "\r\n",
    "dataset.set_split('train')\r\n",
    "train_bar = tqdm(desc='split=train',\r\n",
    "                          total=dataset.get_num_batches(args.batch_size), \r\n",
    "                          position=1, \r\n",
    "                          leave=True)\r\n",
    "dataset.set_split('val')\r\n",
    "val_bar = tqdm(desc='split=val',\r\n",
    "                        total=dataset.get_num_batches(args.batch_size), \r\n",
    "                        position=1, \r\n",
    "                        leave=True)\r\n",
    "\r\n",
    "try:\r\n",
    "    for epoch_index in range(args.num_epochs):\r\n",
    "        train_state['epoch_index'] = epoch_index\r\n",
    "\r\n",
    "        # Iterate over training dataset\r\n",
    "\r\n",
    "        # setup: batch generator, set loss and acc to 0, set train mode on\r\n",
    "        dataset.set_split('train')\r\n",
    "        batch_generator = generate_batches(dataset, \r\n",
    "                                           batch_size=args.batch_size, \r\n",
    "                                           device=args.device)\r\n",
    "        running_loss = 0.0\r\n",
    "        running_acc = 0.0\r\n",
    "        model.train()\r\n",
    "        \r\n",
    "        for batch_index, batch_dict in enumerate(batch_generator):\r\n",
    "            # the training routine is these 5 steps:\r\n",
    "\r\n",
    "            # --------------------------------------    \r\n",
    "            # step 1. zero the gradients\r\n",
    "            optimizer.zero_grad()\r\n",
    "\r\n",
    "            # step 2. compute the output\r\n",
    "            y_pred = model(x_in=batch_dict['x_data'])\r\n",
    "\r\n",
    "            # step 3. compute the loss\r\n",
    "            loss = sequence_loss(y_pred, batch_dict['y_target'], mask_index)\r\n",
    "\r\n",
    "\r\n",
    "            # step 4. use loss to produce gradients\r\n",
    "            loss.backward()\r\n",
    "\r\n",
    "            # step 5. use optimizer to take gradient step\r\n",
    "            optimizer.step()\r\n",
    "            # -----------------------------------------\r\n",
    "            # compute the  running loss and running accuracy\r\n",
    "            running_loss += (loss.item() - running_loss) / (batch_index + 1)\r\n",
    "            acc_t = compute_accuracy(y_pred, batch_dict['y_target'], mask_index)\r\n",
    "            running_acc += (acc_t - running_acc) / (batch_index + 1)\r\n",
    "\r\n",
    "            # update bar\r\n",
    "            train_bar.set_postfix(loss=running_loss,\r\n",
    "                                  acc=running_acc,\r\n",
    "                                  epoch=epoch_index)\r\n",
    "            train_bar.update()\r\n",
    "\r\n",
    "        train_state['train_loss'].append(running_loss)\r\n",
    "        train_state['train_acc'].append(running_acc)\r\n",
    "\r\n",
    "        # Iterate over val dataset\r\n",
    "\r\n",
    "        # setup: batch generator, set loss and acc to 0; set eval mode on\r\n",
    "        dataset.set_split('val')\r\n",
    "        batch_generator = generate_batches(dataset, \r\n",
    "                                           batch_size=args.batch_size, \r\n",
    "                                           device=args.device)\r\n",
    "        running_loss = 0.\r\n",
    "        running_acc = 0.\r\n",
    "        model.eval()\r\n",
    "\r\n",
    "        for batch_index, batch_dict in enumerate(batch_generator):\r\n",
    "            # compute the output\r\n",
    "            y_pred = model(x_in=batch_dict['x_data'])\r\n",
    "\r\n",
    "            # step 3. compute the loss\r\n",
    "            loss = sequence_loss(y_pred, batch_dict['y_target'], mask_index)\r\n",
    "\r\n",
    "            # compute the  running loss and running accuracy\r\n",
    "            running_loss += (loss.item() - running_loss) / (batch_index + 1)\r\n",
    "            acc_t = compute_accuracy(y_pred, batch_dict['y_target'], mask_index)\r\n",
    "            running_acc += (acc_t - running_acc) / (batch_index + 1)\r\n",
    "            \r\n",
    "            # Update bar\r\n",
    "            val_bar.set_postfix(loss=running_loss, acc=running_acc, \r\n",
    "                            epoch=epoch_index)\r\n",
    "            val_bar.update()\r\n",
    "\r\n",
    "        train_state['val_loss'].append(running_loss)\r\n",
    "        train_state['val_acc'].append(running_acc)\r\n",
    "\r\n",
    "        train_state = update_train_state(args=args, model=model, \r\n",
    "                                         train_state=train_state)\r\n",
    "\r\n",
    "        scheduler.step(train_state['val_loss'][-1])\r\n",
    "\r\n",
    "        if train_state['stop_early']:\r\n",
    "            break\r\n",
    "        \r\n",
    "        # move model to cpu for sampling\r\n",
    "        model = model.cpu()\r\n",
    "        sampled_surnames = decode_samples(\r\n",
    "            sample_from_model(model, vectorizer, num_samples=2), \r\n",
    "            vectorizer)\r\n",
    "        epoch_bar.set_postfix(sample1=sampled_surnames[0], \r\n",
    "                              sample2=sampled_surnames[1])\r\n",
    "        # move model back to whichever device it should be on\r\n",
    "        model = model.to(args.device)\r\n",
    "        \r\n",
    "        train_bar.n = 0\r\n",
    "        val_bar.n = 0\r\n",
    "        epoch_bar.update()\r\n",
    "        \r\n",
    "except KeyboardInterrupt:\r\n",
    "    print(\"Exiting loop\")"
   ],
   "outputs": [
    {
     "output_type": "display_data",
     "data": {
      "application/vnd.jupyter.widget-view+json": {
       "version_major": 2,
       "version_minor": 0,
       "model_id": "391d6bd1886b49d3b4a58fe730164392"
      },
      "text/plain": [
       "training routine:   0%|          | 0/100 [00:00<?, ?it/s]"
      ]
     },
     "metadata": {}
    },
    {
     "output_type": "display_data",
     "data": {
      "application/vnd.jupyter.widget-view+json": {
       "version_major": 2,
       "version_minor": 0,
       "model_id": "8057a8bf9b944c7381085bd6e39f453b"
      },
      "text/plain": [
       "split=train:   0%|          | 0/60 [00:00<?, ?it/s]"
      ]
     },
     "metadata": {}
    },
    {
     "output_type": "display_data",
     "data": {
      "application/vnd.jupyter.widget-view+json": {
       "version_major": 2,
       "version_minor": 0,
       "model_id": "368f32dde2814f8882d18108fb66b03f"
      },
      "text/plain": [
       "split=val:   0%|          | 0/12 [00:00<?, ?it/s]"
      ]
     },
     "metadata": {}
    }
   ],
   "metadata": {
    "code_folding": []
   }
  },
  {
   "cell_type": "code",
   "execution_count": 18,
   "source": [
    "np.random.choice(np.arange(len(vectorizer.nationality_vocab)), replace=True, size=2)"
   ],
   "outputs": [
    {
     "output_type": "execute_result",
     "data": {
      "text/plain": [
       "array([8, 7])"
      ]
     },
     "metadata": {},
     "execution_count": 18
    }
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 19,
   "source": [
    "# compute the loss & accuracy on the test set using the best available model\r\n",
    "\r\n",
    "model.load_state_dict(torch.load(train_state['model_filename']))\r\n",
    "\r\n",
    "model = model.to(args.device)\r\n",
    "\r\n",
    "dataset.set_split('test')\r\n",
    "batch_generator = generate_batches(dataset, \r\n",
    "                                   batch_size=args.batch_size, \r\n",
    "                                   device=args.device)\r\n",
    "running_acc = 0.\r\n",
    "model.eval()\r\n",
    "\r\n",
    "for batch_index, batch_dict in enumerate(batch_generator):\r\n",
    "    # compute the output\r\n",
    "    y_pred = model(x_in=batch_dict['x_data'])\r\n",
    "\r\n",
    "    # compute the loss\r\n",
    "    loss = sequence_loss(y_pred, batch_dict['y_target'], mask_index)\r\n",
    "\r\n",
    "    # compute the accuracy\r\n",
    "    running_loss += (loss.item() - running_loss) / (batch_index + 1)\r\n",
    "\r\n",
    "    acc_t = compute_accuracy(y_pred, batch_dict['y_target'], mask_index)\r\n",
    "    running_acc += (acc_t - running_acc) / (batch_index + 1)\r\n",
    "\r\n",
    "train_state['test_loss'] = running_loss \r\n",
    "train_state['test_acc'] = running_acc "
   ],
   "outputs": [],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 15,
   "source": [
    "print(\"Test loss: {};\".format(train_state['test_loss']))\r\n",
    "print(\"Test Accuracy: {}\".format(train_state['test_acc']))"
   ],
   "outputs": [
    {
     "output_type": "stream",
     "name": "stdout",
     "text": [
      "Test loss: 2.6623813112576804;\n",
      "Test Accuracy: 23.699346354510485\n"
     ]
    }
   ],
   "metadata": {}
  },
  {
   "cell_type": "markdown",
   "source": [
    "## Inference"
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 20,
   "source": [
    "# number of names to generate\r\n",
    "num_names = 10\r\n",
    "model = model.cpu()\r\n",
    "# Generate nationality hidden state\r\n",
    "sampled_surnames = decode_samples(\r\n",
    "    sample_from_model(model, vectorizer, num_samples=num_names), \r\n",
    "    vectorizer)\r\n",
    "# Show results\r\n",
    "print (\"-\"*15)\r\n",
    "for i in range(num_names):\r\n",
    "    print (sampled_surnames[i])"
   ],
   "outputs": [
    {
     "output_type": "stream",
     "name": "stdout",
     "text": [
      "---------------\n",
      "Clan\n",
      "Turkuek\n",
      "Srigbo\n",
      "T\n",
      "Gchare\n",
      "Wagil\n",
      "Noélaveyo\n",
      "Faiausmeiduz\n",
      "Spylik\n",
      "Wardhin\n"
     ]
    }
   ],
   "metadata": {}
  },
  {
   "cell_type": "markdown",
   "source": [
    "Using the functions `sample_from_model()` and `decode_samples()`, we can inspect the output of the model above, to get a sense of whether the model is learning to generate sensible surnames. \r\n",
    "\r\n",
    ":::{admonition} What can we learn from inspecting the output? \r\n",
    "We can see that although the surnames appear to follow several morphological patterns, the names don't appear distinctly to be of one nationality or another. One possibility is that learning a general model of surnames confuses the character distributions between different nationalities. The [conditioned `SurnameGenerationModel`](https://weiliu2k.github.io/CITS4012/LSTM/Surname_Generation_Conditioned.html#) is meant to handle this kind of situation."
   ],
   "metadata": {}
  }
 ],
 "metadata": {
  "kernelspec": {
   "name": "python3",
   "display_name": "Python 3.8.10 64-bit ('cits4012': conda)"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.8.10"
  },
  "toc": {
   "colors": {
    "hover_highlight": "#DAA520",
    "running_highlight": "#FF0000",
    "selected_highlight": "#FFD700"
   },
   "moveMenuLeft": true,
   "nav_menu": {
    "height": "264px",
    "width": "252px"
   },
   "navigate_menu": true,
   "number_sections": true,
   "sideBar": true,
   "threshold": "5",
   "toc_cell": false,
   "toc_section_display": "block",
   "toc_window_display": false
  },
  "interpreter": {
   "hash": "d990147e05fc0cc60dd3871899a6233eb6a5324c1885ded43d013dc915f7e535"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}